#!/usr/local/bin/perl -w

# Format.pl

# SPECIFICATIONS

# System (Mac, UNIX, or PC)

$System = 'Mac';

# Directories (input, output)

$InputDirectory = 'Separated';
$OutputDirectory = 'Formatted';

# Division marker delimiters

@Division = (
	'<',	# Open delimiter
	'>'		# Close delimiter
	);

# Tag delimiters

@Tag = (
	'[',	# Start-tag open delimiter
	'[/',	# End-tag open delimiter
	']'		# Tag close delimiter
	);

# Certainty tag generic identifiers

@Certainties = (
	'ut',	# Uncertain text
	'rt'	# Reconstructed text
	);

# Corresponding certainties

$Uncertain = 0.5;
$Reconstructed = 0.05;

# Concatenation marker

$Concatenation = '=';

# Newline character and punctuation. These must not be
# alphanumeric, the underscore (_), or empty ('').
# Punctuation must not include the newline character.

$NewLine = '|';

@Punctuation = (
	".", "?", "?", ",", "?", ";", ":", "?", "'", "-",
	"+", "?"
	);

# Include newline and punctuation characters?
# 0 = no, 1 = yes.

$IncNewLines = 0;
$IncPunctuation = 1;

# END SPECIFICATIONS

# Print introduction

print "Format:\n\n";

print "This program rearranges each witness file into the\n";
print "format required for subsequent steps:\n";
print "(1) eradicate potential problems\n";
print "- discard text preceding first div. marker\n";
print "- replace tabs with spaces\n";
print "(2) process line divisions\n";
print "- discard concatenation markers and associated newlines\n";
print "- discard multiple consecutive newlines\n";
print "- replace remaining newlines with specified character\n";
print "- isolate newline characters\n";
print "(3) process punctuation\n";
print "- isolate specified punctuation\n";
print "(4) process tags\n";
print "- discard all tags except certainty level tags\n";
print "- assign certainty to each character according to tags\n";
print "(5) generate output.\n\n";

print "Notes:\n\n";

print "(1) The newline character and punctuation must not be\n";
print "    alphanumeric characters or the underscore (_).\n";
print "(2) Ensure that specified punctuation is comprehensive.\n";
print "    (The Punctuation utility lists possible punctuation.)\n";
print "    Punctuation must not include the newline character.\n";
print "(3) Each output character is assigned a certainty\n";
print "    (i.e. probability of being what is reported)\n";
print "    according to its certainty level tags:\n";
print "    (a) no tags = certainty of 1.00\n";
print "    (b) uncertain text tags  = as specified\n";
print "    (c) reconstructed text tags  = as specified.\n";
print "    The certainty of each output item is calculated from\n";
print "    the certainties of its characters.\n";
print "(4) Output is formatted as follows:\n";
print "    Item TAB Certainty NEWLINE\n";
print "(5) Newline and punctuation characters must be specified\n";
print "    but may be excluded from output if desired.\n\n";

# Show specifications

print "Specifications:\n\n";

print "$System version.\n\n";

print "Input will be read from the directory named ",
	$InputDirectory, ".\n";
print "Output will be placed in the directory named ",
	$OutputDirectory, ".\n\n";

print "Division markers should look like this:\n",
	$Division[0], "DIVISION MARKER", $Division[1], "\n\n";

print "Certainty tags should look like this:\n";
print $Tag[0], $Certainties[0], $Tag[2], "UNCERTAIN TEXT",
	$Tag[1], $Certainties[0], $Tag[2], "\n";
print $Tag[0], $Certainties[1], $Tag[2], "RECONSTRUCTED TEXT",
	$Tag[1], $Certainties[1], $Tag[2], "\n\n";

print "Certainties are assigned as follows:\n";
print "Uncertain text probability = $Uncertain\n";
print "Reconstructed text probability = $Reconstructed\n\n";

print "Concatenation marker: $Concatenation\n\n";
if($Concatenation eq '')
{ die "Error: The concatenation character has not been specified.\n" }

print "Newline character: $NewLine\n\n";
if($NewLine eq '')
{ die "Error: The newline character has not been specified.\n" }
if($NewLine =~ /\w/) { die "Error: $NewLine is alphanumeric.\n" }

print "Punctuation (ASCII value and character equivalent):\n";
$i = 1;
foreach $Item (sort(@Punctuation))
{
	printf("%2d: %3d = %s\n", $i, ord($Item), $Item);
	if($Item =~ /\w/) { die "Error: $Item is alphanumeric.\n" }
	if($Item eq '')
		{ die "Error: This punctuation specification is empty.\n" }
	if($Item eq $NewLine)
		{ die "Error: Punctuation is newline character.\n" }
	$i += 1;
}

unless($IncNewLines and $IncPunctuation)
{ print "\nFeatures excluded from output:\n" }
unless($IncNewLines) { print "newline characters\n" }
unless($IncPunctuation) { print "punctuation\n" }

# Make output directory

mkdir($OutputDirectory, 0770);

# Select directory character

$System = uc($System);	# Make all upper-case
if($System eq 'MAC') { $Dir = ':' }
elsif($System eq 'UNIX') { $Dir = '/' }
elsif($System eq 'PC') { $Dir = '\\' }
else { die "Error: Directory character for $System not known\n" }

# Make paths to directories

$InputPath = $Dir.$InputDirectory.$Dir;
$OutputPath = $Dir.$OutputDirectory.$Dir;
unless($System eq 'MAC')
{
	$InputPath = '.'.$InputPath;
	$OutputPath = '.'.$OutputPath;
}

# Open input directory

opendir(INDIR, $InputDirectory) or die
	"Can't open input directory $InputDirectory";

# Make input directory contents list

@Contents = readdir(INDIR);

# Remove . and .. from directory list

unless($System eq 'MAC') { splice(@Contents, 0, 2) }

# Make patterns for regular expressions

$UTS = quotemeta($Tag[0].$Certainties[0].$Tag[2]);	# Unc. start
$UTE = quotemeta($Tag[1].$Certainties[0].$Tag[2]);	# Unc. end
$RTS = quotemeta($Tag[0].$Certainties[1].$Tag[2]);	# Rec. start
$RTE = quotemeta($Tag[1].$Certainties[1].$Tag[2]);	# Rec. end
$DMO = quotemeta($Division[0]);	# Division Marker Open
$DMC = quotemeta($Division[1]);	# Division Marker Close

# FORMAT TRANSCRIPTIONS

print "\nFormat transcriptions:\n\n";

foreach $InputName (@Contents)
{
	# Open input file for reading

	open(INPUT, '<'.$InputPath.$InputName) or die
		"Can't open input file $InputPath$InputName";
	print "$InputPath$InputName\n";

	# Make transcription into one string

	$Input = join('', <INPUT>);

	# Close file

	close(INPUT);

	# Eradicate potential problems:

	# (1) Discard text preceding first div. marker

	unless($Input =~ /^$DMO.*?$DMC/)
	{
		$Input =~ s/(^[\S\W]+?)($DMO.*?$DMC)/$2/;
		print "Text preceding first div. marker discarded:\n";
		print "$1\n";
	}

	# (2) Replace tabs with spaces

	if($Input =~ /\t/)
	{
		$Input =~ s/\t/ /g;
		print "Tabs replaced with spaces.\n";
	}

	# (3) Add newline at end

	$Input = $Input."\n";

	# Discard concatenation markers and associated newlines

	$Pattern = quotemeta($Concatenation);
	$Input =~ s/$Pattern\n//g;

	# Discard multiple consecutive newlines

	$Input =~ tr/\n/\n/s;

	# Replace remaining newlines with specified character and
	# isolate if required. Otherwise replace with space.

	if($IncNewLines) { $Input =~ s/\n/ $NewLine /g }
	else { $Input =~ s/\n/ /g }

	# Isolate or exclude punctuation

	if($IncPunctuation)
	{
		foreach $Item (@Punctuation)
		{
			$Pattern = quotemeta($Item);
			$Input =~ s/($Pattern)/ $1 /g;
		}
	}
	else
	{
		foreach $Item (@Punctuation)
		{
			$Pattern = quotemeta($Item);
			$Input =~ s/$Pattern//g;
		}
	}

	# Discard all tags except certainty level tags

	@Identifiers = &ListIdentifiers($Tag[0], $Tag[2], $Input);

	foreach $GI (@Identifiers)
	{
		unless(($GI eq $Certainties[0]) or ($GI eq $Certainties[1]))
		{
			$Start = $Tag[0].$GI.$Tag[2];
			$End = $Tag[1].$GI.$Tag[2];
			print "Discard ", $Start, " and ", $End, " tags.\n";
			$Discard = quotemeta($Start);
			$Input =~ s/$Discard//g;
			$Discard = quotemeta($End);
			$Input =~ s/$Discard//g;
		}
	}

	# Separate text according to certainty levels

	$Input =~ s/($UTS)/\n$1/g;
	$Input =~ s/($UTE)/$1\n/g;
	$Input =~ s/($RTS)/\n$1/g;
	$Input =~ s/($RTE)/$1\n/g;

	# Separate division marker components

	$Input =~ s/($DMO.*?$DMC)/\n$1\n/g;

	# Split separated lines into list

	@List = split(/\n/, $Input);

	# Specify certainty of each text character

	$Output = '';

	foreach $Line (@List)
	{
		# Skip blank lines

		if($Line eq '') { next }

		# Enter division markers

		elsif($Line =~ /$DMO.*?$DMC/)
		{ $Output = $Output.$Line."\n" }

		# Handle uncertain text

		elsif($Line =~ /$UTS(.*)$UTE/)
		{
			$Line = $1;
			$Line =~ s/(.)/$1\t$Uncertain\n/g;
			$Output = $Output.$Line;
		}

		# Handle reconstructed text

		elsif($Line =~ /$RTS(.*)$RTE/)
		{
			$Line = $1;
			$Line =~ s/(.)/$1\t$Reconstructed\n/g;
			$Output = $Output.$Line;
		}

		# Handle certain text

		else
		{
			$Line =~ s/(.)/$1\t1.00\n/g;
			$Output = $Output.$Line;
		}
	}

	# Join div. marker components

	$Output =~ s/($DMC)\n($DMO)/$1$2/g;

	# Split output into list

	@List = split(/\n/, $Output);

	# Build output

	$Output = '';
	$Characters = '';
	$Certainty = 0;

	foreach $Item (@List)
	{
		$Size = length($Characters);

		# Add character to character list and add certainty
		# to sum of certainties if character encountered

		if($Item =~ /([^ ])\t(.+)/)
		{
			$Characters = $Characters.$1;
			$Certainty += $2;
		}

		# Add characters and certainty to output if space
		# encountered and characters accumulated.

		elsif(($Item =~ /^ /) and ($Size > 0))
		{
			$Certainty = $Certainty/$Size;
			$Item = sprintf "%s\t%.3f\n", $Characters, $Certainty;
			$Output = $Output.$Item;

			# Reset characters and certainty

			$Characters = '';
			$Certainty = 0;
		}

		# Add div. marker to output if encountered.

		elsif($Item =~ /$DMO.*$DMC/)
		{ $Output = $Output.$Item."\n" }

		# Skip if space encountered and no characters accumulated.

		elsif(($Item =~ /^ /) and ($Size == 0)) { next }

		else { die "Error: Incorrect input format: $Item\n" }
	}

	# Check for remaining cert. tags

	foreach $Cert (@Certainties)
	{
		$Find = quotemeta($Tag[0].$Cert.$Tag[2]);

		if($Output =~ /($Find.*)/)
		{ die "Error: Unprocessed certainty tag: $1" }

		$Find = quotemeta($Tag[1].$Cert.$Tag[2]);

		if($Output =~ /(.*$Find)/)
		{ die "Error: Unprocessed certainty tag: $1" }
	}


	# Make output file name

	$OutputName = $InputName;

	# Open output file for writing

	open(OUTPUT, '>'.$OutputPath.$OutputName) or die
			"Can't write to file $OutputName in $OutputDirectory";

	print "->  $OutputPath$OutputName\n";

	# Write output to file

	print OUTPUT $Output;

	# Close file

	close(OUTPUT);
}

# Print end message

$Duration = time - $^T;

print "\nFormat finished in $Duration seconds.\n";

sub ListIdentifiers
{
	# Make list of generic identifiers from start tags

	my($Open) = quotemeta($_[0]);
	my($Close) = quotemeta($_[1]);
	my($Input) = $_[2];
	my(@Identifiers) = ();

	push(@Identifiers, $Input =~ /$Open(\w*?)$Close/g);

	# Sort and compact list

	@Identifiers = sort @Identifiers;

	my($Identifier);
	my($Previous) = '';
	my(@List) = ();

	foreach $Identifier (@Identifiers)
	{
		if($Identifier ne $Previous)
		{ push(@List, $Identifier) }

		$Previous = $Identifier;
	}

	return @List;
}
