#!/usr/local/bin/perl -w

# Matrix.pl

# SPECIFICATIONS

# System (Mac, UNIX, or PC)

$System = 'Mac';

# Directories (input, synthesis, unwanted, output)

$InputDirectory = 'Normalised';
$SynthDirectory = 'Synthesised';
$UnwantedDirectory = 'Unwanted';
$OutputDirectory = 'Matrices';

# Files (synthesis, unwanted)

$SynthName = 'Synthesis';
$UnwantedName = 'Textual';

# Division marker delimiters

@Division = (
	'<',	# Open delimiter
	'>'		# Close delimiter
	);

# Threshold. Any item with a certainty below the
# threshold will be classified as a missing datum.

$Threshold = 0.5;

# Features to include. 0 = no, 1 = yes.

$IncLabels = 0;		# Row and column labels
$IncDivs = 0;		# Division markers
$IncSingles = 0;	# Singleton variations
$IncNoVars = 0;		# Rows with no variation
$IncMissing = 0;	# Rows with missing data

# END SPECIFICATIONS

# Print introduction

print "Matrix:\n\n";

print "Build a data matrix for each input file.\n\n";

print "Notes:\n\n";

print "(1) Each data matrix is based on a synthetic sequence\n";
print " containing all sequences found in the witness files.\n";
print " (The Synthesise utility builds this.)\n";
print " The same witnesses as used to build the synthetic\n";
print " sequence must be used here.\n";
print "(2) The synthetic sequence must be properly formatted.\n";
print " (The Synthesise utility does this.)\n";
print " Witness files must also be properly formatted.\n";
print " (Either the Format or Normalise utility does this.)\n";
print "(3) Each data matrix is restricted to the divisions for\n";
print " which its witness file (the principal witness) exists.\n";
print " Other witnesses are only included if they exist for\n";
print " all of the principal witness file divisions.\n";
print "(4) Data matrix entries have the following values:\n";
print " 1 = present\n";
print " 0 = absent\n";
print " -1 = missing data\n";
print " An item is classified as missing data if its\n";
print " certainty is below a specified threshold.\n";
print "(5) Output data matrices consist of one tab separated\n";
print " row vector per principal witness item, with one column\n";
print " per witness.\n";
print "(6) The following data matrix features may be included\n";
print " or excluded as desired:\n";
print " - row and column labels\n";
print " - division markers\n";
print " - rows with a singleton variation\n";
print " - rows with no variation\n";
print " - rows with missing data.\n";
print " Rows corresponding to items listed in a specified\n";
print " unwanted items file will also be excluded.\n";
print "(7) A separate file containing the following items is\n";
print " generated if labels and div. markers are excluded:\n";
print " - column labels with a fixed length of 8 characters\n";
print " - row labels\n";
print " - dimensions of the data matrix.\n\n";

# Select directory character

$System = uc($System);	# Make all upper-case
if($System eq 'MAC') { $Dir = ':' }
elsif($System eq 'UNIX') { $Dir = '/' }
elsif($System eq 'PC') { $Dir = '\\' }
else { die "Error: Directory character for $System not known\n" }

# Make paths to directories

$InputPath = $Dir.$InputDirectory.$Dir;
$SynthPath = $Dir.$SynthDirectory.$Dir;
$UnwantedPath = $Dir.$UnwantedDirectory.$Dir;
$OutputPath = $Dir.$OutputDirectory.$Dir;
unless($System eq 'MAC')
{
	$InputPath = '.'.$InputPath;
	$SynthPath = '.'.$SynthPath;
	$UnwantedPath = '.'.$UnwantedPath;
	$OutputPath = '.'.$OutputPath;
}

# Open input directory

opendir(INDIR, $InputDirectory) or die
	"Can't open input directory $InputDirectory\n";

# Make input directory contents list

@Contents = readdir(INDIR);

# Remove . and .. from directory list

unless($System eq 'MAC') { splice(@Contents, 0, 2) }

# Make output directory

mkdir($OutputDirectory, 0770);

# Show specifications

print "Specifications:\n\n";

print "$System version.\n\n";

print "Input witness files directory: $InputDirectory.\n";
print "Synthetic sequence file: $SynthPath$SynthName.\n";
print "Unwanted items file: $UnwantedPath$UnwantedName.\n";
print "Output matrix files directory: $OutputDirectory.\n\n";

print "Division markers should look like this:\n";
print "$Division[0]DIVISION MARKER$Division[1]\n\n";

print "Any witness item with a certainty below $Threshold\n";
print "will be classified as a missing datum.\n\n";

unless($IncLabels and $IncDivs and $IncSingles and $IncNoVars
	and $IncMissing)
{ print "Excluded data matrix features:\n" }
unless($IncLabels) { print "column and row labels\n" }
unless($IncDivs) { print "division markers\n" }
unless($IncSingles) { print "rows with a singleton variation\n" }
unless($IncNoVars) { print "rows with no variation\n" }
unless($IncMissing) { print "rows with missing data\n" }

if(($IncLabels == 0) and ($IncDivs == 0))
{ print "\nSeparate labels files will be written out.\n" }

# Make patterns for regular expressions

$DMO = quotemeta($Division[0]);
$DMC = quotemeta($Division[1]);

# Read unwanted items file if it exists.

if(open(UNWANTED, '<'.$UnwantedPath.$UnwantedName))
{
	print "\nRows corresponding to items specified in\n",
	"$UnwantedPath$UnwantedName will also be excluded.\n";

	@Unwanted = <UNWANTED>;

	# Close file

	close(UNWANTED);

	# Remove newlines

	chomp(@Unwanted);

	# Check format

	foreach $Item (@Unwanted)
	{
		if($Item eq '')
		{ die "Error: Blank line detected" }
		elsif($Item =~ /\t/)
		{ die "Error: Tab detected: $Item" }
		elsif($Item =~ /$DMO.*$DMC/)
		{ die "Error: Division marker detected: $Item" }
	}
}
else
{
	@Unwanted = ();
	print "\nCan't find specified unwanted items file\n",
	"$UnwantedPath$UnwantedName.\n",
	"Unwanted items will not be discarded.\n";
}

# Open synthesis file for reading

print "\nRead synthesis file $SynthPath$SynthName\n";

open(SYNTH, '<'.$SynthPath.$SynthName) or die
	"Error: Can't open $SynthPath$SynthName";

@InputList = <SYNTH>;

# Close synthesis file

close(SYNTH);

# Remove newlines

chomp(@InputList);

# Check format

foreach $Item (@Unwanted)
{
	if($Item eq '') { die "Error: Blank line detected" }
	elsif($Item =~ /\t/) { die "Error: Tab detected: $Item" }
}

# Convert to string

$Input = join("\n", @InputList);

# Convert synthesis to associative array with divisions as keys

%Synthesis = ();

@InputList = &DivArray($DMO, $DMC, $Input);

for($i = 0; $i < @InputList; $i += 2)
{ $Synthesis{$InputList[$i]} = $InputList[$i+1] }

# BUILD MATRICES

print "\nBuild matrices:\n\n";

foreach $PrincipalName (@Contents)
{
	# Read principal file

	open(PRINCIPAL, '<'.$InputPath.$PrincipalName) or die
		"Can't open principal file $InputPath$PrincipalName";

	print "Principal witness: $InputPath$PrincipalName\n";

	@InputList = <PRINCIPAL>;

	# Close file

	close(PRINCIPAL);

	# Remove newlines

	chomp(@InputList);

	# Check format

	foreach $Item (@InputList)
	{
		if($Item =~ /^.+\t.+$/) { next }
		elsif($Item =~ /^$DMO.+$DMC$/) { next }
		else { die "Error: Incorrect input format: $Item" }
	}

	# Convert to string

	$Input = join("\n", @InputList);

	# Make list of principal divisions

	@Principal = ();

	@InputList = &DivArray($DMO, $DMC, $Input);

	for($i = 0; $i < @InputList; $i += 2)
	{ push(@Principal, $InputList[$i]) }

	# Initialise columns list

	@Columns = ('R/C');

	# Add row labels

	foreach $Div (@Principal)
	{
		# Add div. marker

		{ push(@Columns, $Div) }

		# Add div. items

		@SynthesisList = split(/\n/, $Synthesis{$Div});
		push(@Columns, @SynthesisList);
	}

	# Count rows

	$RowCount = @Columns;

	# Add column for each witness which exists for all principal
	# witness divisions.

	foreach $InputName (@Contents)
	{
		# Read input file

		open(INPUT, '<'.$InputPath.$InputName) or die
			"Can't open input file $InputPath$InputName";

		@InputList = <INPUT>;

		# Close file

		close(INPUT);

		# Convert to associative array with divisions as keys

		%Witness = ();

		$Input = join('', @InputList);

		@InputList = &DivArray($DMO, $DMC, $Input);

		# Enter div. marker as key and text as value

		for($i = 0; $i < @InputList; $i += 2)
		{ $Witness{$InputList[$i]} = $InputList[$i+1] }

		# Test whether principal and witness files have
		# all divisions in common

		$Test = 1;	# 0 = not all common, 1 = all common

		foreach $Div (@Principal)
		{
			unless(exists($Witness{$Div}))
			{
				$Test = 0;
				last;
			}
		}

		# Make witness column if all divisions common

		if($Test)
		{
			print "<- $InputPath$InputName\n";

			# Add column label

			push(@Columns, $InputName);

			# Add entries for each item of each division

			foreach $Div (@Principal)
			{
				# Add blank for div. marker

				push(@Columns, '');

				# Split master and witness div. texts into lists

				@SynthesisList = split(/\n/, $Synthesis{$Div});
				@WitnessList = split(/\n/, $Witness{$Div});

				# Initialise previous certainty

				$PrevCert = 0;

				# Make column entries for witness items

				foreach $W (@WitnessList)
				{
					# Get witness item and certainty

					$W =~ /(.+)\t(.+)/;
					$W = $1;
					$Cert = $2;

					# Set matrix entry

					if($PrevCert < $Threshold) { $Entry = '-1' }
					else { $Entry = '0' }

					until($W eq ($M = shift(@SynthesisList)))
					{
						# Enter 0 if master item not in witness
						# or -1 if previous cert. below threshold

						push(@Columns, $Entry);

						# Die if witness item not in master

						unless(defined($M))
						{ die "Error: $W not in master\n" }
					}

					# Set matrix entry

					if($Cert < $Threshold) { $Entry = '-1' }
					else { $Entry = '1' }

					# Enter 1 if master item in witness
					# or -1 if cert. below threshold

					push(@Columns, $Entry);

					# Update previous cert.

					$PrevCert = $Cert;
				}

				# Add column entries for remaining master items

				# Set matrix entry

				if($PrevCert < $Threshold) { $Entry = '-1' }
				else { $Entry = '0' }

				while(@SynthesisList)
				{
					push(@Columns, $Entry);
					shift(@SynthesisList);
				}
			}
		}
	}

	# Make output file name

	$OutputName = $PrincipalName;

	# Open output file for writing

	open(OUTPUT, '>'.$OutputPath.$OutputName) or die
		"Can't write to file $OutputPath$OutputDirectory";

	print "-> $OutputPath$OutputName\n";

	# Make column labels list

	@ColumnLabels = ();

	$i = 0;
	for($j = 0; $j < @Columns; $j += $RowCount)
	{ push(@ColumnLabels, $Columns[$i + $j]) }

	# Write out column labels if required

	if($IncLabels)
	{ print OUTPUT join("\t", @ColumnLabels), "\n" }

	# Shift first entry

	shift(@ColumnLabels);

	# Write out matrix rows as required

	@RowLabels = ();

	for($i = 1; $i < $RowCount; $i += 1)
	{
		# Get row

		@Row = ();

		for($j = 0; $j < @Columns; $j += $RowCount)
		{ push(@Row, $Columns[$i + $j]) }

		# Skip div. marker if excluded

		unless($IncDivs)
		{ if($Row[0] =~ /^$DMO.+$DMC$/) { next } }

		# Skip unwanted item if excluded

		if(@Unwanted)
		{
			$Pattern = quotemeta($Row[0]);
			if(grep(/^$Pattern$/, @Unwanted))
			{ next }
		}

		# Shift row label out of row

		$RowLabel = shift(@Row);

		# Count witnesses with item absent (coded by 0)

		$Absent = grep(/0/, @Row);

		# Count witnesses with item present (coded by 1)

		$Present = grep(/^1/, @Row);

		# Count witnesses with missing datum (coded by -1)

		$Missing = grep(/^-1/, @Row);

		# Count witnesses which could have item

		$Possible = $Present + $Missing;

		# Skip singleton if excluded

		unless($IncSingles)
		{ if(($Absent == 1) or ($Possible == 1)) { next } }

		# Skip rows with no variation if excluded

		unless($IncNoVars)
		{ if(($Absent == 0) or ($Possible == 0)) { next } }

		# Skip rows with missing data if excluded

		unless($IncMissing)
		{ if($Missing) { next } }

		# Add row label to row labels list

		push(@RowLabels, $RowLabel);

		# Write out row label if required

		if($IncLabels) { print OUTPUT $RowLabel, "\t" }

		# Write out rest of row

		print OUTPUT join("\t", @Row), "\n";
	}

	# Close file

	close(OUTPUT);

	# Write out labels file if required

	if(($IncLabels == 0) and ($IncDivs == 0))
	{
		# Make labels file name

		$OutputName = $PrincipalName."_labels";

		# Open labels file for writing

		open(LABELS, '>'.$OutputPath.$OutputName) or die
		"Can't write to file $OutputPath$OutputName";

		print "-> $OutputPath$OutputName\n";

		# Format and write out column labels.
		# Each label has 8 characters.
		# There is a maximum of 10 labels per line.

		$i = 0;
		foreach $Item (@ColumnLabels)
		{
			$Remainder = $i % 10;
			if(($i) and ($Remainder == 0)) { print LABELS "\n" }
			printf LABELS "%-8s", $Item;
			$i += 1;
		}

		# Write out row labels

		print LABELS "\n\nRow Labels:\n", join("\t", @RowLabels);

		# Add matrix dimensions

		$C = @ColumnLabels;
		$R = @RowLabels;

		print LABELS
		"\n\nMatrix dimensions: $R rows x $C columns.\n";

		# Close file

		close(LABELS);
	}

	print "\n";
}

# Print end message

$Duration = time - $^T;

print "Matrix finished in $Duration seconds.\n";

sub DivArray
{
	# Convert input to list of div. markers and associated texts.
	# First two items of input list are div. markers with
	# metacharacters quoted.
	# Input assumed to be list of items separated by newlines.
	# Each div. marker must be first item on its line.
	# Extraneous characters following div. marker are returned.

	my($DMO) = $_[0];
	my($DMC) = $_[1];
	my(@InputList) = split(/\n/, $_[2]);
	my($Item);
	my(@List) = ();
	my($Text) = '';

	# (Input reversed to facilitate separation of divisions)

	foreach $Item (reverse(@InputList))
	{
		# Update list if div. marker encountered

		if($Item =~ /^$DMO.+$DMC$/)
		{
			$Text =~ s/\n$//;	# Remove last newline from text
			push(@List, $Text);	# Add text to list
			push(@List, $Item);	# Add div. marker to list
			$Text = '';			# Reset text
		}

		# Otherwise update text

		else { $Text = $Item."\n".$Text }
	}

	# Reverse list again to restore original order and return

	return reverse(@List);
}
