#!/usr/local/bin/perl -w

# Normalise.pl

# SPECIFICATIONS

# System (Mac, UNIX, or PC)

$System = 'Mac';

# Directories (input, table, output)

$InputDirectory = 'Formatted';
$TableDirectory = 'Equivalents';
$OutputDirectory = 'Normalised';

# File (table)

$TableName = 'Hebrews';

# Division marker delimiters

@Division = (
	'<',	# Open delimiter
	'>'		# Close delimiter
	);

# Newline character

$NewLine = '|';

# Special characters to be treated as alphabetical

@Alphabetical = ( "?", "" );

# END SPECIFICATIONS

# Print introduction

print "Normalise:\n\n";

print "This program normalises witness files according to a\n";
print "specified equivalents table.\n\n";

print "Notes:\n\n";

print "(1) Witness files must be properly formatted.\n";
print " (The Format utility does this.)\n";
print "(2) Each record (i.e. row) of the equivalents table must\n";
print " have the following format:\n";
print " Pattern TAB Replacement ... NEWLINE\n";
print " The table may contain other tab-separated fields.\n";
print "(3) Patterns and replacements may be space-separated\n";
print " phrases.\n";
print "(4) Non-alphabetical items in the pattern or replacement\n";
print " will be isolated from preceding text.\n";
print "(5) Ensure that special characters to be treated as\n";
print " alphabetical are specified.\n";
print "(6) The certainty of each replacement phrase item is the\n";
print " least of the pattern phrase items\' certainties.\n";
print "(7) Normalised witness file records are output in the\n";
print " same format as input witness file records:\n";
print " Item TAB Certainty NEWLINE\n\n";

# Show specifications

print "Specifications:\n\n";

print "$System version.\n\n";

print "Input witness files directory: $InputDirectory.\n";
print "Equivalents table directory: $TableDirectory.\n";
print "Equivalents table name: $TableName.\n";
print "Output witness files directory: $OutputDirectory.\n\n";

print "Division markers should look like this:\n",
	"$Division[0]DIVISION MARKER$Division[1]\n\n";

print "Newline character:\n$NewLine\n\n";
if($NewLine eq '')
{ die "Error: The newline character has not been specified" }

print "Special characters to be treated as alphabetical:\n";
foreach $Item (@Alphabetical) { print "$Item\n" }

# Select directory character

$System = uc($System);	# Make all upper-case
if($System eq 'MAC') { $Dir = ':' }
elsif($System eq 'UNIX') { $Dir = '/' }
elsif($System eq 'PC') { $Dir = '\\' }
else { die "Error: Directory character for $System not known" }

# Make paths to directories

$InputPath = $Dir.$InputDirectory.$Dir;
$TablePath = $Dir.$TableDirectory.$Dir;
$OutputPath = $Dir.$OutputDirectory.$Dir;
unless($System eq 'MAC')
{
	$InputPath = '.'.$InputPath;
	$TablePath = '.'.$TablePath;
	$OutputPath = '.'.$OutputPath;
}

# Open input directory

opendir(INDIR, $InputDirectory) or die
	"Can't open input directory $InputDirectory";

# Make input directory contents list

@Contents = readdir(INDIR);

# Remove . and .. from directory list

unless($System eq 'MAC') { splice(@Contents, 0, 2) }

# Make output directory

mkdir($OutputDirectory, 0770);

# Make patterns for regular expressions

$Special = quotemeta(join("", @Alphabetical));
$DMO = quotemeta($Division[0]);
$DMC = quotemeta($Division[1]);

# Open table file for reading

print "\nRead equivalents table\n\n";

open(TABLE, '<'.$TablePath.$TableName) or die
"Error: Can't open table file $TablePath$TableName";

@Table = <TABLE>;

# Close table file

close(TABLE);

# Remove newline from each record

chomp(@Table);

# Convert table file to arrays

@Simple = ();	# For simple substitutions
@Complex = ();	# For complex substitutions

foreach $Record (@Table)
{
	@Fields = split(/\t/, $Record);
	
	if(@Fields < 2)
	{ die "Error: Pattern or replacement missing: $Record" }
	
	$Pattern = $Fields[0];
	$Replacement = $Fields[1];
	
	# Space out non-alphabetical characters
	
	$Pattern =~
		s/([a-zA-Z$Special])([^a-zA-Z$Special])/$1 $2 /g;
	$Replacement =~
		s/([a-zA-Z$Special])([^a-zA-Z$Special])/$1 $2 /g;
	
	# Discard multiple spaces
	
	$Pattern =~ tr/ / /s;
	$Replacement =~ tr/ / /s;
	
	# Add to appropriate array
	
	if(($Pattern !~ / /) and ($Replacement !~ / /))
	{ push(@Simple, $Pattern, $Replacement) }
	else
	{ push(@Complex, $Pattern, $Replacement) }
}

# NORMALISE TRANSCRIPTIONS

print "Normalise transcriptions:\n\n";

foreach $InputName (@Contents)
{
	# Read input file
	
	open(INPUT, '<'.$InputPath.$InputName) or die
		"Can't open input file $InputPath$InputName";
	
	print "$InputPath$InputName\n";
	
	@InputList = <INPUT>;
	
	# Close file
	
	close(INPUT);
	
	# Check format
	
	unless($InputList[0] =~ /^$DMO.+$DMC\n$/)
	{ die "Error: First item not a div. marker: $InputList[0]" }
	
	unless($InputList[1] =~ /^.+\t(.+)\n$/)
	{ die "Error: Incorrect input format: $InputList[1]" }
	$CertSize = length($1);
	
	foreach $Item (@InputList)
	{
		if($Item =~ /^.+\t(.+)\n$/)
		{
			unless(length($1) == $CertSize)
			{ die "Error: Differing certainty format: $1" }
		}
		elsif($Item =~ /^$DMO.+$DMC\n$/) { next }
		else { die "Error: Incorrect input format: $Item" }
	}
	
	# Convert to string
	
	$Output = join("", @InputList);
	@InputList = ();
	
	# Add initial newline
	
	$Output = "\n".$Output;
	
	# Make complex substitutions
	
	for($i = 0; $i < @Complex; $i += 2)
	{
		# Initialise pattern and replacement arrays
		
		@Pattern = split(" ", $Complex[$i]);
		@Replacement = split(" ", $Complex[$i+1]);
		
		# Turn on printing switch
		
		$Switch = 1;
		
		# Locate start of pattern
		
		$Index = index($Output, "\n$Pattern[0]\t", 0);
		
		while($Index != -1)
		{
			# print "Index: $Index\n";
			
			$Index += 1;	# Move past initial newline
			$Start = $Index;	# Set start
			$Match = 1;
			@Certainties = ();
			@NewLines = ();
			
			# Check for match of each item in pattern
			
			foreach $Item (@Pattern)
			{
				$Size = length($Item);
				$Segment = substr($Output, $Index, $Size);
				$Next = substr($Output, $Index + $Size, 1);
				# print "Item: $Item\n";
				# print "Segment: $Segment\n";
				# print "Next character: $Next\n";
				
				# Check for match
				
				if(($Segment eq $Item) and ($Next eq "\t"))
				{
					# Get certainty and add to list
					
					$Cert =
					substr($Output, $Index + $Size + 1, $CertSize);
					# print "Certainty: $Cert\n";
					push(@Certainties, $Cert);
					
					# Move index to first character of next segment
					
					$Index += $Size + $CertSize + 2;
					
					# Handle newline character.
					
					if(substr($Output, $Index, 1) eq $NewLine)
					{
						push(@NewLines, $NewLine);	# Add to list
						$Index += $CertSize + 3;	# Move index
						if(substr($Output, $Index, 1) eq $NewLine)
						{ die "Error: Multiple newlines detected" }
					}
					else { push(@NewLines, '') }
					
					# print "Next: ",substr($Output,$Index,1),"\n";
				}
				else
				{
					$Match = 0;
					last;
				}
			}
			
			# Perform substitution if match detected
			
			if($Match == 1)
			{
				# Print substitution once
				
				if($Switch)
				{
					# print "$Complex[$i] -> $Complex[$i+1]\n";
					$Switch = 0;	# Turn off printing switch
				}
				
				# Get least certainty
				
				$Cert = $Certainties[0];
				foreach $Item (@Certainties)
				{ if($Item < $Cert) { $Cert = $Item } }
				
				# Make parts for output string
				
				$First = substr($Output, 0, $Start);
				
				$Middle = "";
				
				foreach $Item (@Replacement)
				{
					# Add replacement
					
					$Segment = $Item."\t".$Cert."\n";
					$Middle = $Middle.$Segment;
					
					# Add newline character if required
					
					if(@NewLines)
					{
						if(shift(@NewLines) eq $NewLine)
						{
							$Segment = $NewLine."\t".$Cert."\n";
							$Middle = $Middle.$Segment;
						}
					}
				}
				
				# Add final newline character if required
				
				if(@NewLines)
				{
					if(pop(@NewLines) eq $NewLine)
					{
						$Segment = $NewLine."\t".$Cert."\n";
						$Middle = $Middle.$Segment;
					}
				}
				
				$Remainder = length($Output) - $Index;
				$Last = substr($Output, $Index, $Remainder);
				
				# Make substitution
				
				$Output = $First.$Middle.$Last;
				
				# Recalculate index
				
				$Index = $Start + length($Middle);
			}
			
			# Locate next start of pattern
			
			$Index = index($Output, "\n$Pattern[0]\t", $Index);
		}
	}
	
	# Discard consecutive newline characters
	
	$Pattern = quotemeta($NewLine);
	$Output =~ s/$Pattern\t.+?\n$Pattern/$NewLine/g;
	
	# Make simple substitutions
	
	for($i = 0; $i < @Simple; $i += 2)
	{
		$Pattern = quotemeta($Simple[$i]);
		$Replacement = $Simple[$i+1];
		if($Output =~ /\n$Pattern\t/)
		{
			# print "$Simple[$i] -> $Simple[$i+1]\n";
			$Output =~ s/\n$Pattern\t/\n$Replacement\t/g;
		}
	}
	
	# Remove initial newline
	
	$Output =~ s/^\n//;
	
	# Make output file name
	
	$OutputName = $InputName;
	
	# Open output file for writing
	
	open(OUTPUT, '>'.$OutputPath.$OutputName) or die
		"Can't write to file $OutputPath$OutputDirectory";
			
	print "-> $OutputPath$OutputName\n\n";
	
	# Write output to file
	
	print OUTPUT $Output;
	
	# Close file
	
	close(OUTPUT);
}

# Print end message

$Duration = time - $^T;

print "Normalise finished in $Duration seconds.\n";
