#!/usr/local/bin/perl -w

# Report.pl

# SPECIFICATIONS

# System (Mac, UNIX, or PC)

$System = 'Mac';

# Directories (input, synthesis, output)

$InputDirectory = 'Normalised';
$SynthDirectory = 'Synthesised';
$OutputDirectory = 'Report';

# Files (synthesis, output)

$SynthName = 'Synthesis';
$OutputName = 'Report';

# Division marker delimiters

@Division = (
	'<',	# Open delimiter
	'>'		# Close delimiter
	);

# Threshold. Any item with a certainty below the
# threshold will be classified as a missing datum.

$Threshold = 0.5;

# END SPECIFICATIONS

# Print introduction

print "Report:\n\n";

print "Produce a concise report of input texts.\n\n";

print "Notes:\n\n";

print "(1) The report is based on a synthetic sequence which\n";
print "    contains every sequence of items found in the\n";
print "    input texts.\n";
print "(2) Entries have the following values:\n";
print "    * = present\n";
print "    - = absent\n";
print "    ? = missing data\n";
print "    An item is classified as missing data if its\n";
print "    certainty is below a specified threshold.\n\n";

# Select directory character

$System = uc($System);	# Make all upper-case
if($System eq 'MAC') { $Dir = ':' }
elsif($System eq 'UNIX') { $Dir = '/' }
elsif($System eq 'PC') { $Dir = '\\' }
else { die "Error: Directory character for $System not known\n" }

# Make paths to directories

$InputPath = $Dir.$InputDirectory.$Dir;
$SynthPath = $Dir.$SynthDirectory.$Dir;
$OutputPath = $Dir.$OutputDirectory.$Dir;
unless($System eq 'MAC')
{
	$InputPath = '.'.$InputPath;
	$SynthPath = '.'.$SynthPath;
	$OutputPath = '.'.$OutputPath;
}

# Open input directory

opendir(INDIR, $InputDirectory) or die
	"Can't open input directory $InputDirectory\n";

# Make input directory contents list

@Contents = readdir(INDIR);

# Remove . and .. from directory list

unless($System eq 'MAC') { splice(@Contents, 0, 2) }

# Make output directory

mkdir($OutputDirectory, 0770);

# Show specifications

print "Specifications:\n\n";

print "$System version.\n\n";

print "Input directory: $InputDirectory.\n";
print "Synthetic sequence file: $SynthPath$SynthName.\n";
print "Output file: $OutputPath$OutputName.\n\n";

print "Division markers should look like this:\n";
print "$Division[0]DIVISION MARKER$Division[1]\n\n";
	
print "Any items with a certainty below $Threshold\n";
print "will be classified as missing data.\n\n";

# Make patterns for regular expressions

$DMO = quotemeta($Division[0]);
$DMC = quotemeta($Division[1]);

# Read synthetic sequence file

print "Read synthetic sequence file $SynthPath$SynthName\n\n";

open(SYNTH, '<'.$SynthPath.$SynthName) or die
	"Error: Can't open $SynthPath$SynthName";

@Input = <SYNTH>;

# Close file

close(SYNTH);

# Remove newlines

chomp(@Input);

# Check format

foreach $Line (@Input)
{
	if($Line eq '') { die "Error: Blank line detected" }
	elsif($Line =~ /\t/) { die "Error: Tab detected: $Line" }
}

# Convert to string

$String = join("\n", @Input);

# Convert to synthetic sequence associative array.
# Div. is key, div. + text is value.

%Synth = ();

@Input = &DivArray($DMO, $DMC, $String);

for($i = 0; $i < @Input; $i += 2)
{ $Synth{$Input[$i]} = $Input[$i+1] }

# Initialise report

%Report = ();

# MAKE REPORT

print "Make report:\n\n";

foreach $File (@Contents)
{
	# Read input file
	
	open(IN, "<$InputPath$File") or die
	"Can't read $InputPath$File";
	
	print "<- $InputPath$File\n";
	
	@Input = <IN>;
	
	# Close file
	
	close(IN);
	
	# Remove newlines
	
	chomp(@Input);
	
	# Check format
	
	foreach $Line (@Input)
	{
		if($Line =~ /^.+\t.+$/) { next }
		elsif($Line =~ /^$DMO.+$DMC$/) { next }
		else { die "Error: Incorrect input format: $Line" }
	}
	
	# Convert to string
	
	$String = join("\n", @Input);
	
	# Make list of divisions and texts
	
	@List = &DivArray($DMO, $DMC, $String);
	
	# Report text for each div.
	
	while(@List)
	{
		$Div = shift(@List);
#		print "$Div\n";
		
		# Get synthetic sequence for div.
		
		$MS = $Synth{$Div};
		
		# Get witness sequence for div.
		
		$WS = shift(@List);
		
		# Make report record
		
		$Record = &MakeRecord($MS, $WS, $Threshold, $Div);
		
		# Add record to report
		
		if(exists($Report{$Div}))
		{ $Report{$Div} = $Report{$Div}.$File."\n".$Record."\n" }
		else
		{ $Report{$Div} = $File."\n".$Record."\n" }
	}
}

# Open output file

open(OUTPUT, ">$OutputPath$OutputName") or die
"Can't write $OutputPath$OutputName";

# Write output
			
print "\n-> $OutputPath$OutputName\n";

foreach $Div (sort DivSort keys(%Report))
{
	# Get list of elements and matrix dimensions.

	@List = ($Div);								# Add div. marker
	push(@List, split(/\n/, $Synth{$Div}));		# Add items
	$Rows = @List;								# Number of rows
	push(@List, split(/\n/, $Report{$Div}));	# Add elements
	$Total = @List;
	
	# Write out rows
	
	for($i = 0; $i < $Rows; $i += 1)
	{
		@Row = ();
		
		for($j = 0; $j < $Total; $j += $Rows)
		{ push(@Row, $List[$i+$j]) }
		
		# Write out row unless all entries absent
		
		$Absent = grep(/^-/, @Row);
		
		unless($Absent == @Row - 1)
		{ print OUTPUT join("\t", @Row), "\n" }
	}
}

# Print end message

$Duration = time - $^T;

print "\nReport finished in $Duration seconds.\n";

sub DivArray
{
	# Convert input to list of div. markers and associated texts.
	# First two items of input list must be div. markers with
	# metacharacters quoted.
	# Input assumed to be list of items separated by newlines.
	# Each div. marker must be only item on its line.
	
	my($DMO) = $_[0];
	my($DMC) = $_[1];
	my(@Input) = split(/\n/, $_[2]);
	my($Item);
	my(@List) = ();
	my($Text) = '';
	
	# (Input reversed to facilitate separation of divisions)
	
	foreach $Item (reverse(@Input))
	{
		# Update list if div. marker encountered
		
		if($Item =~ /^$DMO.+$DMC$/)
		{
			$Text =~ s/\n$//;	# Remove last newline from text
			push(@List, $Text);	# Add text to list
			push(@List, $Item);	# Add div. marker to  list
			$Text = '';			# Reset text
		}
		
		# Otherwise update text
		
		else { $Text = $Item."\n".$Text }
	}
	
	# Reverse list again to restore original order and return
	
	return reverse(@List);
}

sub MakeRecord
{
	# Make a record from a primary and secondary sequence.
	# The primary sequence string must have the following format:
	# Item NEWLINE Item NEWLINE ...
	# The secondary sequence string must have the following format:
	# Item TAB Certainty NEWLINE Item TAB Certainty NEWLINE ...
	# The primary sequence forms the basis of the record.
	# Entries have the following values:
	# * = present
	# - = absent
	# ? = missing data (i.e. certainty below threshold).
	
	my(@PS) = split(/\n/, $_[0]);
	my(@SS) = split(/\n/, $_[1]);
	my($Threshold) = $_[2];
	my($Div) = $_[3];
	my($p, $Line, $s, $Cert);
	my($PrevCert) = 0;
	my(@List) = ();
	
#	$Cert = @PS;
#	print $Cert, "\t", join("\t", @PS), "\n";
	
	while (@SS)
	{
		# Initialise primary item
		
		unless(defined($p = shift(@PS)))
		{ die "Error: Primary sequence insufficient at $Div" }
		
		# Get secondary item and cert.
		
		$Line = shift(@SS);
		unless($Line =~ /^(.+)\t(.+)$/)
		{ die "Error: Incorrect input format: $Line" }
		$s = $1;
		$Cert = $2;
		
		# Set entry
		
		if($PrevCert < $Threshold) { $Entry = '?' }
		else {  $Entry = '-' }
		
		until($p eq $s)
		{
			# Enter - if absent
			# or ? if prev. cert. below threshold
			
			push(@List, $Entry);
			
			unless(defined($p = shift(@PS)))
			{ die "Error: Primary sequence insufficient at $Div" }
		}
		
		# Set entry
		
		if($Cert < $Threshold) { $Entry = '?' }
		else {  $Entry = '*' }
		
		# Enter * if present
		# or ? if cert. below threshold
		
		push(@List, $Entry);
		
		# Update prev. cert.
		
		$PrevCert = $Cert;
	}
	
	# Complete record
	
	# Set entry
	
	if($PrevCert < $Threshold) { $Entry = '?' }
	else {  $Entry = '-' }
	
	while(@PS)
	{
		push(@List, $Entry);
		shift(@PS);
	}
	
	# Return record
	
#	$Cert = @List;
#	print $Cert, "\t", join("\t", @List), "\n";
	
	return join("\n", @List);
}

sub DivSort
{
	# Sort div. markers
	
	my($a1, $a2, $b1, $b2, $i);
	
	unless($a =~ /.+?(\d+).+?(\d+)/)
	{ die "Bad division format: $a" }
	$a1 = $1;
	$a2 = $2;

	unless($b =~ /.+?(\d+).+?(\d+)/)
	{ die "Bad division format: $b" }
	$b1 = $1;
	$b2 = $2;
	
	if($a1 > $b1) { $i = 1 }
	if($a1 == $b1)
	{
		if($a2 > $b2) {$i = 1 }
		if($a2 == $b2) {$i = 0 }
		if($a2 < $b2) {$i = -1 }
	}
	if($a1 < $b1) { $i = -1 }
	
	return $i;
}
