#!/usr/local/bin/perl -w

# Separate.pl

# SPECIFICATIONS

# System (Mac, UNIX, or PC)

$System = 'Mac';

# Directories (input, output)

$InputDirectory = 'Filtered';
$OutputDirectory = 'Separated';

# Division marker delimiters

@Division = (
	'<',	# Open delimiter
	'>'		# Close delimiter
	);

# Tag delimiters

@Tag = (
	'[',	# Start-tag open delimiter
	'[/',	# End-tag open delimiter
	'd',	# Deletion code
	'c',	# Insertion code
	']'		# Tag close delimiter
	);

# END SPECIFICATIONS

# Print introduction

print "Separate: Separates scribal stages.\n\n";

print "Notes:\n\n";

print "Alteration tag generic identifiers must contain:\n";
print "(a) an alphanumeric insertion or deletion code, then\n";
print "(b) a scribal identification number (0,1,2,...,x).\n";
print " x denotes alteration by an unidentified scribe.\n";
print "The original scribe stage is separated for all divisions.\n";
print "Subsequent stages are separated for altered divisions alone.\n\n";

# Show specifications

print "Specifications:\n\n";

print "$System version.\n\n";

print "Input directory: $InputDirectory\n";
print "Output directory: $OutputDirectory\n\n";

# Make division marker components

$DMO = $Division[0];	# Division Marker Open
$DMC = $Division[1];	# Division Marker Close

print "Division markers should look like this:\n",
	$DMO, "DIVISION MARKER", $DMC, "\n\n";

# Make tag components
	
$DTSO = $Tag[0].$Tag[2];	# Deletion Tag Start Open
$DTEO = $Tag[1].$Tag[2];	# Deletion Tag End Open
$ITSO = $Tag[0].$Tag[3];	# Insertion Tag Start Open
$ITEO = $Tag[1].$Tag[3];	# Insertion Tag End Open
$TC = $Tag[4];				# Tag Close
	
print "Alteration tags should look like this (n = 0,1,2,...,x):\n",
	$DTSO, "n", $TC, "DELETION", $DTEO, "n", $TC, " ",
	$ITSO, "n", $TC, "INSERTION", $ITEO, "n", $TC, "\n\n";

# Make patterns for regular expressions
	
$DMO = quotemeta($DMO);
$DMC = quotemeta($DMC);
$DTSO = quotemeta($DTSO);
$DTEO = quotemeta($DTEO);
$ITSO = quotemeta($ITSO);
$ITEO = quotemeta($ITEO);
$TC = quotemeta($TC);

# Make output directory

mkdir($OutputDirectory, 0770);

# Select directory character

$System = uc($System);	# Make all upper-case
if($System eq 'MAC') { $Dir = ':' }
elsif($System eq 'UNIX') { $Dir = '/' }
elsif($System eq 'PC') { $Dir = '\\' }
else { die "Error: Directory character for $System not known\n" }

# Make paths to directories

$InputPath = $Dir.$InputDirectory.$Dir;
$OutputPath = $Dir.$OutputDirectory.$Dir;
unless($System eq 'MAC')
{
	$InputPath = '.'.$InputPath;
	$OutputPath = '.'.$OutputPath;
}

# Open input directory

opendir(INDIR, $InputDirectory) or die
	"Can't open input directory $InputDirectory";

# Make input directory contents list

@Contents = readdir(INDIR);

# Remove . and .. from directory list

unless($System eq 'MAC') { splice(@Contents, 0, 2) }

# SEPARATE SCRIBAL STAGES

print "Separate scribal stages:\n";

foreach $InputName (@Contents)
{
	# Open input file for reading
	
	open(INPUT, '<'.$InputPath.$InputName) or die
		"Can't open input file $InputPath$InputName\n";
		
	print $InputPath, $InputName, "\n";
	
	# Make transcription into one string
	
	@InputList = <INPUT>;
	$Input = join('', @InputList);

	# Close file
	
	close(INPUT);
	
	# Make list of divisions

	@Divisions = &ListDivisions($Input);
	
	# Make list of scribes
		
	@Scribes = &ListScribes($Input);
	
	# SEPARATE ORIGINAL STAGE FOR ALL DIVISIONS
	
	# Make string of relevant divisions
	
	$Original = $Input;
	
	# Make lists of earlier and later scribes
	
	@Earlier = ();
	@Later = @Scribes;
	
	# Exclude later alterations
	
	$Original = &ExcludeEarlier($Original, @Later);
	
	# Make output file name
	
#	if(@Scribes > 0) { $OutputName = $InputName.'-*' }
#	else { $OutputName = $InputName }
	
	$OutputName = $InputName;
	
	# Open output file for writing
	
	open(OUTPUT, '>'.$OutputPath.$OutputName) or die
			"Can't write to file $OutputName in $OutputDirectory";
	print "-> ", $OutputPath.$OutputName, "\n";
	
	# Write output to file
	
	print OUTPUT $Original;
	
	# Close file
	
	close(OUTPUT);
	
	# SEPARATE SUBSEQUENT STAGES FOR ALTERED DIVISIONS ALONE
	
	foreach $Scribe (@Scribes)
	{
		# Make string of relevant divisions
		
		@Relevant = grep(/[$DTSO$ITSO]$Scribe$TC/, @Divisions);
		$Subsequent = join('', @Relevant);
		
		# Update lists of earlier and later scribes
		
		push(@Earlier, shift(@Later));
		
		# Exclude later alterations
		
		$Subsequent = &ExcludeEarlier($Subsequent, @Later);
		
		# Include earlier alterations
		
		$Subsequent = &IncludeEarlier($Subsequent, @Earlier);
		
		# Make output file name (1st corr. = 1, 2nd corr. = 2, ...)
		
		$OutputName = $InputName.'-'.$Scribe;

		# Open output file for writing
		
		open(OUTPUT, '>'.$OutputPath.$OutputName) or die
			"Can't write to file $OutputName in $OutputDirectory";
		print "-> ", $OutputPath.$OutputName, "\n";
		
		# Print out result
	
		print OUTPUT $Subsequent;
		
		# Close file
	
		close(OUTPUT);
	}
}

# Print end message
	
$Duration = time - $^T;

print "\nSeparate finished in $Duration seconds.\n";

sub ListDivisions	# Make list of divisions
{
	my($Input) = $_[0];
	my(@Parts) = ();
	my(@List) = ();
	my($Item);
	my($Division) = "";
	
	@Parts = split(/($DMO[\S\W]*?$DMC)/, $Input);
	
	foreach $Item (@Parts)
	{
		if($Item =~ /$DMO[\S\W]*?$DMC/)
		{ $Division = $Division.$Item }
		elsif($Item =~ /[\S\W]+/)
		{
			$Division = $Division.$Item;
			push(@List, $Division);
			$Division = "";
		}
	}
	
	return @List;
}

sub ListScribes # Make compact list of scribes
{
	# Make list of scribal identifiers
	
	my(@Identifiers) = ();
	push(@Identifiers, $_[0] =~ /$DTSO.*?(\d+|x)$TC/g);
	push(@Identifiers, $_[0] =~ /$ITSO.*?(\d+|x)$TC/g);
	
	# Sort and compact list
	
	@Identifiers = sort @Identifiers;
	
	my($Identifier) = '';
	my($Previous) = '';
	my(@List) = ();
	
	foreach $Identifier (@Identifiers)
	{
		if($Identifier ne $Previous)
		{ push(@List, $Identifier) }
		
		$Previous = $Identifier;
	}
	
	return @List;
}

sub ExcludeEarlier	# Exclude alterations by later scribes
{
	my($Output) = $_[0];
	my($i);
	
	for($i = 1; $i <= $#_; $i += 1)
	{
		# Delete later deletion tags
		
		$Output =~ s/$DTSO$_[$i]$TC([\S\W]*?)$DTEO$_[$i]$TC/$1/g;
		
		# Delete later insertions and associated tags
		
		$Output =~ s/$ITSO$_[$i]$TC[\S\W]*?$ITEO$_[$i]$TC//g;
	}
	
	return $Output;
}

sub IncludeEarlier	# Include alterations by earlier scribes
{
	my($Output) = $_[0];
	my($i);
	
	for($i = 1; $i <= $#_; $i += 1)
	{
		# Delete earlier deletions and associated tags
		
		$Output =~ s/$DTSO$_[$i]$TC[\S\W]*?$DTEO$_[$i]$TC//g;
		
		# Insert earlier insertions and delete associated tags
		
		$Output =~ s/$ITSO$_[$i]$TC([\S\W]*?)$ITEO$_[$i]$TC/$1/g;
	}
	
	return $Output;
}
