#!/usr/local/bin/perl -w

# Synthesise.pl

# SPECIFICATIONS

# System (Mac, UNIX, or PC)

$System = 'Mac';

# Directories (input, synthesis)

$InputDirectory = 'Normalised';
$SynthDirectory = 'Synthesised';

# File (synthesis)

$SynthName = 'Synthesis';

# Division marker delimiters

@Division = (
	'<',	# Open delimiter
	'>'		# Close delimiter
	);

# END SPECIFICATIONS

# Print introduction

print "Synthesise:\n\n";

print "Synthesise a sequence from a set of witnesses so\n";
print "that each witness can be reproduced by selecting\n";
print "sequential items from the synthetic sequence.\n\n";

print "Notes:\n\n";

print "(1) Witness files must be properly formatted.\n";
print "    (The Format or Normalise utility does this.)\n\n";

# Select directory character

$System = uc($System);	# Make all upper-case
if($System eq 'MAC') { $Dir = ':' }
elsif($System eq 'UNIX') { $Dir = '/' }
elsif($System eq 'PC') { $Dir = '\\' }
else { die "Error: Directory character for $System not known\n" }

# Make paths to directories

$InputPath = $Dir.$InputDirectory.$Dir;
$OutputPath = $Dir.$SynthDirectory.$Dir;
unless($System eq 'MAC')
{
	$InputPath = '.'.$InputPath;
	$OutputPath = '.'.$OutputPath;
}

# Open input directory

opendir(INDIR, $InputDirectory) or die
	"Can't open input directory $InputDirectory";

# Make input directory contents list

@Contents = readdir(INDIR);

# Remove . and .. from directory list

unless($System eq 'MAC') { splice(@Contents, 0, 2) }

# Show specifications

print "Specifications:\n\n";

print "$System version.\n\n";

print "Witnesses directory: $InputDirectory.\n";
print "Synthetic sequence file: $OutputPath$SynthName.\n\n";

print "Division markers should look like this:\n",
	"$Division[0]DIVISION MARKER$Division[1]\n\n";

# SYNTHESISE PRIMARY SEQUENCE FROM WITNESSES

%Synthesis = ();
$DMO = quotemeta($Division[0]);
$DMC = quotemeta($Division[1]);

print "Synthesise:\n\n";

foreach $InputName (@Contents)
{
	# Read input file
	
	open(INPUT, "<$InputPath$InputName") or die
		"Can't open input file $InputPath$InputName";
	
	print "<- $InputPath$InputName\n";
	
	@InputList = <INPUT>;
	
	# Close file
	
	close(INPUT);
	
	# Remove newlines from input
	
	chomp(@InputList);
	
	# Convert input to list of div. markers and associated texts
	# (Input reversed to facilitate separation of divisions)
	
	@List = ();
	$Text = '';
	
	foreach $Item (reverse(@InputList))
	{
		# Update text if text encountered
		
		if($Item =~ /^(.+)\t.+$/) { $Text = $1."\n".$Text }
		
		# Update list if div. marker encountered
		
		elsif($Item =~ /^$DMO.+$DMC$/)
		{
			$Text =~ s/\n$//;	# Remove last newline from text
			push(@List, $Text);	# Add text to list
			push(@List, $Item);	# Add div. marker to  list
			$Text = '';			# Reset text
		}
		
		else { die "Error: Incorrect input format: $Item" }
	}
	
	# Reverse list again to restore original order
	
	@List = reverse(@List);
	
	# Build synthetic sequence for each witness division
	
	for($i = 0; $i < @List; $i += 2)
	{
		$Div = $List[$i];
		$Text = $List[$i+1];
		
#		print "$Div\n";
		
		if(exists($Synthesis{$Div}))
		{ $Synthesis{$Div} = &Synthesise($Synthesis{$Div}, $Text) }
		else { $Synthesis{$Div} = $Text }
	}
}

# Make output directory

mkdir($SynthDirectory, 0770);

# Open output file for writing

open(OUTPUT, '>'.$OutputPath.$SynthName) or die
	"Can't write to file $OutputPath$SynthDirectory";

print "\n-> $OutputPath$SynthName\n\n";

# Write output to file

foreach $Div (sort DivSort keys(%Synthesis))
{ print OUTPUT "$Div\n$Synthesis{$Div}\n"  }

# Close file

close(OUTPUT);

# Print end message

$Duration = time - $^T;

print "Synthesise finished in $Duration seconds.\n";

sub DivSort
{
	unless($a =~ /.+?(\d+).+?(\d+)/)
	{ die "Bad division format: $a" }
	$a1 = $1;
	$a2 = $2;
	
	unless($b =~ /.+?(\d+).+?(\d+)/)
	{ die "Bad division format: $b" }
	$b1 = $1;
	$b2 = $2;
	# print "b1: $b1 b2: $b2\n";
	
	if($a1 > $b1) { $i = 1 }
	if($a1 == $b1)
	{
		if($a2 > $b2) {$i = 1 }
		if($a2 == $b2) {$i = 0 }
		if($a2 < $b2) {$i = -1 }
	}
	if($a1 < $b1) { $i = -1 }
	
	return $i;
}

sub Synthesise
{
	# Synthesise a new sequence (NS) containing both a primary
	# sequence (PS) and secondary sequence (SS).
	# Input PS, SS strings must be formatted as follows:
	# Item NEWLINE Item NEWLINE ...
	# The NS string is returned in the same format.
	
	my(@PS) = split(/\n/, $_[0]);
	my(@SS) = split(/\n/, $_[1]);
	
	# Check whether SS contained in PS
	
	my($Test, $s, $p);
	
	$Test = 1;
	
	while(@SS)
	{
		# Get next SS and PS items
		
		$s = shift(@SS);
		
		unless(defined($p = shift(@PS)))
		{
			$Test = 0;
			last;
		}
		
#		print "s: $s\n";
#		print "p: $p\n";
		
		# Scroll through PS until SS item matched
		
		until($p eq $s)
		{
			unless(defined($p = shift(@PS)))
			{
				$Test = 0;
				last;
			}
			
#			print "p: $p\n";
		}

		# Exit loop if SS not in PS
		
		unless($Test) { last }
	}
	
	# Return PS if SS in PS
	
	if($Test) { return $_[0] }
	
	# Synthesise NS if SS not in PS
	
#	print "Synthesise\n";
	
	@PS = split(/\n/, $_[0]);
	@SS = split(/\n/, $_[1]);
	my(@IPS, @ISS);
	my(@NS) = ();
	my($Pattern);
	
	while(@SS)
	{
		# Split off initial PS (IPS) and initial SS (ISS)
		# Division is made at common item that is not first in SS.
		
#		print "PS: ", join(' ', @PS), "\n";
#		print "SS: ", join(' ', @SS), "\n";
		
		@IPS = ();
		@ISS = (shift(@SS));	# Shift first SS item to ISS
		
		$Test = 0;
		while(@SS)
		{
			$Find = quotemeta($SS[0]);
			if(grep(/^$Find$/, @PS))
			{
#				print "Match: $SS[0]\n";
				
				$Test = 1;
				until($PS[0] eq $SS[0]) { push(@IPS, shift(@PS)) }
			}
			else
			{
				$Test = 0;
				push(@ISS, shift(@SS));
			}
			
			if($Test) { last }
		}
		
		# Make IPS equal to PS if no match
		
		unless($Test) { @IPS = @PS }
				
		# Synthesise NS from ISS and IPS
		
#		print "IPS: ", join(' ', @IPS), "\n";
#		print "ISS: ", join(' ', @ISS), "\n";
		
		while(@ISS)
		{
			# If first ISS item matched in IPS, output IPS to NS
			# until match reached, then discard match.
			
			$Find = quotemeta($ISS[0]);
			if(grep(/^$Find$/, @IPS))
			{
				until($IPS[0] eq $ISS[0])
				{
#					print "PS -> $IPS[0]\n";
					push(@NS, shift(@IPS));
				}
				
				# Discard match
				
				shift(@IPS);
			}
			
			# Shift first ISS item to NS
			
#			print "SS -> $ISS[0]\n";
			push(@NS, shift(@ISS));
		}
		
		# Join remaining IPS to PS
		
		unshift(@PS, @IPS);
		
		# Exit loop if no match
		
		unless($Test) { last }
	}
	
	# Output remaining IPS to NS
		
	while(@IPS)
	{
#		print "IPS -> $IPS[0]\n";
		push(@NS, shift(@IPS));
	}
	
	# Return new sequence
	
	return join("\n", @NS);
}
