#!/usr/local/bin/perl -w

# CT.pl

# SPECIFICATIONS

# System (Mac, UNIX, or PC)

$System = 'Mac';

# Directories (data and output)

$Data = 'Data';
$Output = 'Constructed';

# Data files (standard text, variation units key, and data matrix)

$ST = 'Standard';
$VK = 'Key';
$DM = 'Matrix';

# Division markers (open and close)

$DMO = '<';
$DMC = '>';

# Reconstructed text markers (open and close)

$RTO = '[rt]';
$RTC = '[/rt]';

# Minimum number of variation units required before text is output

$Minimum = 5;

# END OF SPECIFICATIONS

# Print introduction

print "CT:\n\n";

print "Construct texts from a standard text using a variation\n";
print "units key and a multistate data matrix.\n\n";

print "Notes:\n\n";

print "(1) Texts are constructed by replacing pattern phrases\n";
print "    in the standard text with phrases from the key\n";
print "    according to the data matrix entries.\n";
print "(2) The key consists of a list of variation units.\n";
print "    Each variation unit must be formatted as follows:\n";
print "    Division label\n";
print "    Pattern\n";
print "    First variant\n";
print "    ...\n";
print "    Last variant\n";
print "(3) Variation unit division labels must be formatted as\n";
print "    follows: a.bc\n";
print "    where a is the chapter number\n";
print "          b is the verse number\n";
print "    and   c is an optional letter.\n";
print "(4) The data matrix must be formatted as follows:\n";
print "    R/C W1  W2  ... Wp\n";
print "    V1  D11 D12 ... D1p\n";
print "    V2  D21 D22 ... D2p\n";
print "    ... ... ... Dxy ...\n";
print "    Vn  Dn1 Dn2 ... Dnp\n";
print "    where Vs are variation unit division labels\n";
print "          Ws are witnesses\n";
print "    and   Ds are multistate data entries.\n";
print "    Each D is either a numeral representing a variant or\n";
print "    a negative one (-1) representing a missing datum.\n";
print "(5) Replacements are made as follows:\n";
print "    Dxy  Replacement\n";
print "    0    Pattern\n";
print "    1    First variant\n";
print "    2    Second variant\n";
print "    ...  ...\n";
print "(6) Output texts are only constructed for divisions\n";
print "    in which they are defined by the data matrix.\n";
print "    They are tagged as reconstructed text everywhere\n";
print "    except where their texts are defined.\n";
print "(7) Output texts are only constructed for witnesses\n";
print "    that are defined in at least the specified number\n";
print "    of variation units.\n\n";

# Select directory character

$System = uc($System);	# Make all upper-case
if($System eq 'MAC') { $Dir = ':' }
elsif($System eq 'UNIX') { $Dir = '/' }
elsif($System eq 'PC') { $Dir = '\\' }
else { die "Error: Directory character for $System not known\n" }

# Make paths to directories

$DPath = $Dir.$Data.$Dir;
$OPath = $Dir.$Output.$Dir;
unless($System eq 'MAC')
{
	$DPath = ".$DPath";
	$OPath = ".$OPath";
}

# Print specifications

print "Specifications:\n\n";

print "System: $System\n\n";

print "Standard text:       $DPath$ST\n";
print "Variation units key: $DPath$VK\n";
print "Data matrix:         $DPath$DM\n\n";

print "Standard text division markers must look like this:\n";
print $DMO, "ch a", $DMC, $DMO, "v b", $DMC, "\n\n";

print "Reconstructed text will look like this:\n";
print $RTO, "RECONSTRUCTED", $RTC, "\n\n";

print "A text must be defined in at least $Minimum variation\n";
print "units before it is written to the output.\n\n";

# Make hash array of standard

print "Read standard text\n";

open(STAN, "$DPath$ST") or die "Can't open $DPath$ST";

@List = <STAN>;

close <STAN>;

$Standard = join('', @List);

$Standard =~ s/^[\S\W]*?$DMO/$DMO/;
$Standard =~ s/$DMO/\n$DMO/g;
$Standard =~ s/^\n+//;
$Standard =~ s/$DMC/$DMC\n/g;
$Standard =~ s/\n+$//;
$Standard =~ tr/\n/\n/s;

@List = split(/\n/, $Standard);

%Standard = ();

for($i = 0; $i < @List; $i += 3)
{
	$Div = $List[$i].$List[$i+1];
	unless($Div =~ /^$DMO.+?\d+$DMC$DMO.+?\d+$DMC$/)
	{ die "Bad division format: $Div" }
	$Text = $List[$i+2];
	$Text =~ s/^ //;
	$Text =~ s/ $//;
	$Standard{$Div} = $RTO.$Text.$RTC;
}

# @Divs = sort DivSort (keys(%Standard));
# foreach $Div (@Divs) { print "$Div $Standard{$Div}\n" }

# Make hash array of variants

print "Read variation units key\n";

open(VAR, "$DPath$VK") or die "Can't open $DPath$VK";

@List = <VAR>;

close <VAR>;

chomp(@List);
@List = reverse(@List);

%VA = ();
@Variants = ();

foreach $Item (@List)
{
	if($Item =~ /^\d+?\.\d+?[a-z]*$/)
	{
		$VA{$Item} = [ reverse(@Variants) ];		
		@Variants = ();
	}
	else
	{ push(@Variants, $Item) }
}

# @Labels = sort(keys(%VA));
# foreach $Label (@Labels)
# { print "$Label\n", join("\n", @{$VA{$Label}}), "\n" }

# Read data matrix

print "Read data matrix\n";

open(MAT, "$DPath$DM") or die "Can't open $DPath$DM";

@Matrix = <MAT>;

close <MAT>;

chomp(@Matrix);

@Names = split(/\s+/, shift(@Matrix));	# Text names
shift(@Names);	# Remove blank

@DM = ();
@Labels = ();
$Rows = 0;
$Columns = @Names;

foreach $Record (@Matrix)
{
	@Fields = split(/\s+/, $Record);
	$Label = shift(@Fields);
	push(@Labels, $Label);
	unless(@Fields == $Columns)
	{ die "Error: Wrong number of data matrix columns at $Label" }
	push(@DM, [@Fields]);
	$Rows += 1;
}

print "Dimensions: $Rows r x $Columns c\n";

# Make output directory

mkdir("$Output", 0770);

# Construct texts

print "\nConstruct texts:\n";

for($j = 0; $j < $Columns; $j += 1)
{
	print "$Names[$j]\n";
	
	%Text = ();
	$Counter = 0;
	
	for($i = 0; $i < $Rows; $i += 1)
	{
		$State = $DM[$i][$j];
		if($State eq '-1') { next }
		unless($State =~ /^\d$/)
		{ die "Error: Unrecognisable matrix entry: $State" }
		$Counter += 1;
		
		$Label = $Labels[$i];
		unless(defined($VA{$Label}[$State]))
		{ die "Error: State $State undefined: $Names[$j], $Label" }
		$Pattern = quotemeta($VA{$Label}[0]);
		$Replacement = "$RTC $VA{$Label}[$State] $RTO";
#		print "P: $Pattern\n";
#		print "R: $Replacement\n";
		
		$Label =~ /(\d+)\.(\d+)/;
		$Div = $DMO.'ch '.$1.$DMC.$DMO.'v '.$2.$DMC;
		
		unless(exists($Text{$Div}))
		{ $Text{$Div} = $Standard{$Div} }
		
		$String = $Text{$Div};
		$String =~ s/$Pattern/$Replacement/;
		$Pattern = quotemeta("$RTO$RTC");
		$String =~ s/$Pattern//;
		$Text{$Div} = $String;
	}
	
	if($Counter < $Minimum)
	{
		print "Less than $Minimum v.u.s\n";
		next;
	}
	
	print "$Counter v.u.s -> $OPath$Names[$j]\n";
	
	open(OUT, ">$OPath$Names[$j]") or die
	"Can't open $OPath$Names[$j]";
	
	foreach $Div (sort DivSort (keys(%Text)))
	{ print OUT "$Div $Text{$Div}\n" }
	
	close(OUT);
}

# Print end message
	
$Duration = time - $^T;

print "\nCT finished in $Duration seconds.\n";

sub DivSort
{
	# print "a: $a\n";
	unless($a =~ /^$DMO.+?(\d+)$DMC$DMO.+?(\d+)$DMC$/)
	{ die "Bad division format: $a" }
	$a1 = $1;
	$a2 = $2;
	# print "a1: $a1 a2: $a2\n";
	
	# print "b: $b\n";
	unless($b =~ /^$DMO.+?(\d+)$DMC$DMO.+?(\d+)$DMC$/)
	{ die "Bad division format: $b" }
	$b1 = $1;
	$b2 = $2;
	# print "b1: $b1 b2: $b2\n";
	
	if($a1 > $b1) { $i = 1 }
	if($a1 == $b1)
	{
		if($a2 > $b2) {$i = 1 }
		if($a2 == $b2) {$i = 0 }
		if($a2 < $b2) {$i = -1 }
	}
	if($a1 < $b1) { $i = -1 }
	
	# print "$i\n";
	return $i;
}
